use std::str::FromStr;

use base64::Engine;
use http::header::{ACCEPT, CONTENT_TYPE};
use rand::{rngs::OsRng, Rng};
pub use secrecy::ExposeSecret;
use secrecy::SecretString;
use sha2::Digest;
use sha2::Sha256;
use url::Url;

use crate::{
    algorithms::link_rel,
    http::{Body, CONTENT_TYPE_JSON},
    traits::as_string_or_list,
};

#[derive(thiserror::Error, Debug, miette::Diagnostic)]
pub enum Error {
    #[error("No metadata endpoint could be found.")]
    NoMetadataEndpoint,

    #[error("The metadata endpoint did not declare itself as JSON but as {content_type:?}")]
    MetadataContentTypeInvalid { content_type: String },

    #[error("Failed to parse the JSON of the metadata endpoint.")]
    ParseMetadataJson(#[source] serde_json::Error),

    #[error("Client IDs using IndieAuth have to be a URL.")]
    ClientIdMustBeUrl,

    #[error("A scope can't be an empty string.")]
    EmptyScope,

    #[error("The size of the raw challenge does not comply with the specification.")]
    ChallengeOutOfBounds(usize),
}

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, serde::Serialize, serde::Deserialize)]
pub struct Scope(String);

#[derive(Clone, Debug)]
pub struct AccessToken(SecretString);

impl std::ops::Deref for AccessToken {
    type Target = SecretString;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl AccessToken {
    pub fn new(token: impl ToString) -> Self {
        Self(SecretString::new(token.to_string()))
    }
}

impl Scope {
    pub fn new(scope: impl ToString) -> Result<Self, crate::Error> {
        let scope_str = scope.to_string();
        if scope_str.is_empty() {
            return Err(Error::EmptyScope.into());
        }

        Ok(Self(scope_str))
    }
}

impl From<String> for Scope {
    fn from(scope: String) -> Self {
        Self(scope)
    }
}

impl std::ops::Deref for Scope {
    type Target = String;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

/// A representation of a list of [scopes][Scope].
///
/// This structure allows for the normalization of common behaviors when working
/// with scopes.
///
/// # Examples
/// ```
/// # use indieweb::standards::indieauth::Scopes;
/// # use std::str::FromStr;
/// #
/// let scopes = Scopes::from_str("read create:read").unwrap();
///
/// assert!(scopes.has("read"),
///         "can report if it has a scope");
///
/// assert_eq!(scopes.assert("aircraft"),
///            Err(indieweb::Error::MissingScope(Scopes::from_str("aircraft").unwrap())),
///            "can report if it does not have a scope");
///
/// assert_eq!(scopes.to_string(),
///            "read create:read".to_string(),
///            "easy to pass as a string");
/// ```
#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct Scopes(Vec<Scope>);

impl std::fmt::Display for Scopes {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str(
            &self
                .0
                .iter()
                .map(|s| s.to_string())
                .collect::<Vec<_>>()
                .join(" "),
        )
    }
}

impl FromStr for Scopes {
    type Err = crate::Error;

    fn from_str(scopes_string: &str) -> Result<Self, Self::Err> {
        scopes_string
            .split(' ')
            .try_fold(Vec::default(), |mut acc, scope_str| {
                acc.push(Scope::new(scope_str)?);
                Ok(acc)
            })
            .map(Self)
    }
}

impl Scopes {
    /// Reports if this has no scopes defined.
    ///
    /// ```
    /// # use indieweb::standards::indieauth::Scopes;
    /// # use std::str::FromStr;
    /// assert_eq!(
    ///     Ok(false),
    ///     Scopes::from_str("read")
    ///         .map(|scopes| scopes.is_empty()),
    ///     "there's at least one scope in here"
    ///     )
    /// ```
    pub fn is_empty(&self) -> bool {
        self.0.is_empty()
    }

    /// Confirms if there's a scope matching or beginning to match with the provided value.
    ///
    /// This works with matching a direct scope (like 'create') or a scope with prefixes in
    /// its authored order (like 'update:note'). There's no real "format" to what a scope
    /// can look like, that's completely up to your implementation. For example, capitalist
    /// companies like [Slack](https://en.wikipedia.org/wiki/Slack_(software)) use a combination
    /// of colons and dots for scopes (like "files.metadata:read") and monopolies like
    /// [Google](https://en.wikipedia.org/wiki/Google) use URLs as scopes.
    ///
    /// # Examples
    ///
    /// ```
    /// # use indieweb::standards::indieauth::Scopes;
    /// # use std::str::FromStr;
    /// #
    /// let scopes = Scopes::from_str("read update:note").unwrap();
    /// assert!(scopes.has("read"));
    /// assert!(scopes.has("update:note"));
    /// ```
    pub fn has(&self, scope: &str) -> bool {
        self.0.iter().any(|s| s.to_string().starts_with(scope))
    }

    /// Returns the scopes as a list.
    pub fn as_vec(&self) -> &Vec<Scope> {
        &self.0
    }

    /// Confirm the presence of a provided scope.
    ///
    /// # Examples
    ///
    /// ```
    /// # use indieweb::standards::indieauth::Scopes;
    /// # use std::str::FromStr;
    /// #
    /// let scopes = "read create:note".parse::<Scopes>();
    ///
    /// assert_eq!(
    ///     Ok(()),
    ///     scopes.as_ref().unwrap().assert("read"),
    ///     "confirms explicit scope match without prefix");
    /// assert_eq!(
    ///     Ok(()),
    ///     scopes.as_ref().unwrap().assert("create:note"),
    ///     "confirms explicit scope match with prefix");
    /// assert_eq!(
    ///     Err(indieweb::Error::MissingScope(Scopes::from_str("airplane").unwrap())),
    ///     scopes.as_ref().unwrap().assert("create:note airplane"),
    ///     "denies provided scopes if one is not valid");

    /// ```
    pub fn assert(&self, needed_scopes: &str) -> Result<(), crate::Error> {
        tracing::debug!("Asserting that {:?} exists in {:?}", needed_scopes, self);
        let scope_list: Scopes = needed_scopes.parse().unwrap_or_else(|_| Scopes::default());

        let missing_scopes: Scopes = scope_list
            .as_vec()
            .iter()
            .filter(|expected_scope| !self.has(expected_scope))
            .cloned()
            .collect::<Vec<_>>()
            .into();

        if missing_scopes.is_empty() {
            Ok(())
        } else {
            Err(crate::Error::MissingScope(missing_scopes))
        }
    }

    /// Returns a list of scopes that are matching the list of scopes; a union of the two lists.
    ///
    /// # Examples
    ///
    /// ```
    /// # use indieweb::standards::indieauth::Scopes;
    /// # use std::str::FromStr;
    ///
    /// let wanted_scopes = Scopes::from_str("read update:note").unwrap();
    /// let scopes = Scopes::from_str("read create update:note").unwrap();
    /// assert_eq!(scopes.matching("read update:note"), wanted_scopes);
    /// ```
    pub fn matching(&self, wanted_scopes: &str) -> Self {
        wanted_scopes
            .parse::<Self>()
            .unwrap_or_default()
            .as_vec()
            .iter()
            .filter(|expected_scope| self.has(expected_scope))
            .cloned()
            .collect::<Scopes>()
    }

    pub fn minimal() -> Self {
        Self(vec![Scope("read".to_string())])
    }
}

impl From<Vec<String>> for Scopes {
    fn from(scopes: Vec<String>) -> Self {
        scopes.into_iter().map(Scope::from).collect()
    }
}

impl FromIterator<Scope> for Scopes {
    fn from_iter<T: IntoIterator<Item = Scope>>(iter: T) -> Self {
        Self(iter.into_iter().collect::<Vec<_>>())
    }
}

impl From<Vec<Scope>> for Scopes {
    fn from(scopes: Vec<Scope>) -> Self {
        Self(scopes)
    }
}

impl From<Scopes> for Vec<Scope> {
    fn from(val: Scopes) -> Self {
        val.0
    }
}

impl serde::Serialize for Scopes {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        serializer.serialize_str(&self.to_string())
    }
}

impl<'de> serde::de::Deserialize<'de> for Scopes {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de>,
    {
        as_string_or_list::deserialize(deserializer)?
            .into_iter()
            .filter(|s: &String| !s.is_empty())
            .collect::<Vec<_>>()
            .join(" ")
            .parse()
            .map_err(serde::de::Error::custom)
    }
}

#[derive(Clone, serde::Serialize, serde::Deserialize, Debug, PartialEq, Eq)]
pub enum CodeChallengeMethod {
    S256,
    S512,
    Plain,
}

impl CodeChallengeMethod {
    pub fn recommended() -> Self {
        Self::S256
    }
}

#[allow(clippy::to_string_trait_impl)]
impl ToString for CodeChallengeMethod {
    fn to_string(&self) -> String {
        match self {
            Self::S256 => "S256".to_string(),
            Self::S512 => "S512".to_string(),
            Self::Plain => "PLAIN".to_string(),
        }
    }
}

#[derive(Clone, serde::Serialize, serde::Deserialize, Debug, PartialEq, Eq)]
pub struct CodeChallenge(String);

impl std::ops::Deref for CodeChallenge {
    type Target = String;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl CodeChallenge {
    pub fn new(challenge: impl ToString) -> Result<Self, crate::Error> {
        let challenge_str = challenge.to_string();

        if challenge_str.len() > 128 || challenge_str.len() < 43 {
            return Err(Error::ChallengeOutOfBounds(challenge_str.len()).into());
        }

        // FIXME: Run regular expression against contents.

        Ok(Self(challenge_str))
    }

    pub fn generate(method: CodeChallengeMethod) -> (Self, CodeChallengeMethod) {
        let mut bytes = [0u8; 64];
        OsRng.fill(&mut bytes[..]);
        let hash = match method {
            CodeChallengeMethod::S256 => {
                let mut hasher = Sha256::new();
                hasher.update(bytes);
                hasher.finalize()
            }
            CodeChallengeMethod::S512 => todo!(),
            CodeChallengeMethod::Plain => todo!(),
        };

        (
            CodeChallenge(base64::engine::general_purpose::URL_SAFE_NO_PAD.encode(hash)),
            method,
        )
    }
}

pub struct RequestFields {
    pub client_id: ClientId,
    pub redirect_uri: Url,
    pub state: String,
    pub challenge: CodeChallenge,
    pub challenge_method: CodeChallengeMethod,
    pub scope: Scopes,
}

impl RequestFields {
    pub fn new(
        client_id: &ClientId,
        redirect_uri: &Url,
        state: impl ToString,
    ) -> Result<Self, crate::Error> {
        let (code_challenge, code_challenge_method) =
            CodeChallenge::generate(CodeChallengeMethod::recommended());

        Ok(Self {
            client_id: client_id.clone(),
            redirect_uri: redirect_uri.clone(),
            state: state.to_string(),
            challenge: code_challenge,
            challenge_method: code_challenge_method,
            scope: Default::default(),
        })
    }
}

#[derive(serde::Deserialize, serde::Serialize, Debug, Clone, PartialEq, Eq)]
pub struct ServerMetadata {
    pub issuer: Url,
    pub authorization_endpoint: Url,
    pub token_endpoint: Url,

    #[serde(default, skip_serializing_if = "Option::is_none")]
    pub introspection_endpoint: Option<Url>,

    #[serde(default = "ServerMetadata::recommended_code_challenge_methods")]
    pub code_challenge_methods_supported: Vec<String>,

    #[serde(default, skip_serializing_if = "Scopes::is_empty")]
    pub scopes_supported: Scopes,
}

impl ServerMetadata {
    pub fn recommended_code_challenge_methods() -> Vec<String> {
        vec!["S256".to_string()]
    }

    pub fn new_authorization_request_url(
        &self,
        request: RequestFields,
    ) -> Result<Url, crate::Error> {
        let mut auth_url = self.authorization_endpoint.clone();

        let mut fields = auth_url.query_pairs_mut();
        fields
            .append_pair("response_type", "code")
            .append_pair("client_id", &request.client_id)
            .append_pair("redirect_uri", request.redirect_uri.as_str())
            .append_pair("state", &request.state)
            .append_pair("code_challenge", &request.challenge)
            .append_pair(
                "code_challenge_method",
                &request.challenge_method.to_string(),
            );

        if !request.scope.is_empty() {
            fields.append_pair("scope", &request.scope.to_string());
        }

        // TODO: Support additional fields.
        fields.finish();
        drop(fields);

        Ok(auth_url)
    }
}

#[derive(Debug, Clone)]
pub struct ClientId(String);

impl ClientId {
    pub fn new(client_id: impl ToString) -> Result<Self, crate::Error> {
        let client_id_str = client_id.to_string();
        // TODO: Check if it's a HTTP-based URI, according to the spec.
        if client_id_str.parse::<Url>().is_ok() {
            Ok(Self(client_id_str))
        } else {
            Err(Error::ClientIdMustBeUrl.into())
        }
    }
}

impl std::ops::Deref for ClientId {
    type Target = String;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

pub struct Client<HttpClient: crate::http::Client> {
    pub id: ClientId,
    client: HttpClient,
}

impl<HttpClient: crate::http::Client> Client<HttpClient> {
    const LINK_REL: &'static str = "indieauth-metadata";
    pub fn new(id: impl ToString, http_client: HttpClient) -> Result<Self, crate::Error> {
        Ok(Self {
            id: ClientId(id.to_string()),
            client: http_client,
        })
    }

    pub async fn obtain_metadata(&self, remote_url: &Url) -> Result<ServerMetadata, crate::Error> {
        let mut rels = link_rel::for_url(&self.client, remote_url, &[Self::LINK_REL], "HEAD")
            .await?
            .get(Self::LINK_REL)
            .cloned()
            .unwrap_or_default();

        if rels.is_empty() {
            rels = link_rel::for_url(&self.client, remote_url, &[Self::LINK_REL], "GET")
                .await?
                .get(Self::LINK_REL)
                .cloned()
                .unwrap_or_default();
        }

        let metadata_endpoint_url = if let Some(rel) = rels.first().cloned() {
            rel
        } else {
            // FIXME: Check the `.well-known` endpoint first as well.
            // FIXME: Add flag to check for 'classic' IndieAuth support via authn and token
            // endpoint.
            return Err(Error::NoMetadataEndpoint.into());
        };

        let req = crate::http::Request::builder()
            .uri(metadata_endpoint_url.as_str())
            .header(ACCEPT, "application/json")
            .body(Body::Empty)?;

        let resp = self
            .client
            .send_request(req)
            .await?
            .map(|b| b.as_bytes().to_vec());

        let content_type_header_value = resp
            .headers()
            .get(CONTENT_TYPE)
            .map(|hv| hv.to_str().unwrap_or_default().to_string())
            .unwrap_or_default();

        if !content_type_header_value.starts_with(CONTENT_TYPE_JSON) {
            return Err(Error::MetadataContentTypeInvalid {
                content_type: content_type_header_value,
            }
            .into());
        }

        serde_json::from_slice(&resp.into_body()).map_err(|e| Error::ParseMetadataJson(e).into())
    }
}

#[derive(serde::Serialize, serde::Deserialize, Debug, Clone, PartialEq, Eq)]
pub struct AuthorizationRequestPayload {
    pub client_id: Url,
    #[serde(default)]
    pub redirect_uri: Option<Url>,
    pub state: String,
    #[serde(rename = "code_challenge")]
    pub challenge: CodeChallenge,
    #[serde(rename = "code_challenge_method")]
    pub challenge_method: CodeChallengeMethod,
    #[serde(default = "Scopes::minimal")]
    pub scope: Scopes,
}

mod test;
